import base64
from cStringIO import StringIO
import itertools
import mimetools
import mimetypes
import os
import tempfile
import urllib
import urllib2
import zipfile

import c4d
from c4d import gui, bitmaps, storage, plugins


PLUGIN_ID = 13377331

def get_resources_dir():
    script_dir, script_file = os.path.split(__file__)
    return os.path.join(script_dir, "res") 
    
def get_write_dir():
    path = os.path.join(os.path.join(c4d.storage.GeGetStartupWritePath(), "plugins"), "renderbuzz")
    if not os.path.exists(path):
        os.makedirs(path)
    return path

def create_tmp_dir():
    return tempfile.mkdtemp(prefix="rb")

def get_files(source_dir):
    tmp_files = []
    relroot = os.path.abspath(os.path.join(source_dir, os.pardir))
    for root, dirs, files in os.walk(source_dir):
        for file in files:
            filename = os.path.join(root, file)
            tmp_files.append(filename)
            if os.path.isfile(filename): # regular files only
                os.path.join(os.path.relpath(root, relroot), file)
    return tmp_files

def zip_files(files, target_dir, archive_name):
    zipf = None
    os.chdir(target_dir)
    archive_name = '{0}.zip'.format(archive_name)
    try:
        zipf = zipfile.ZipFile(archive_name, 'w', zipfile.ZIP_DEFLATED)
        
        for f in files:
            if os.path.isfile(f): # regular files only
                arcname = os.path.join(os.path.relpath(os.path.dirname(f),
                                       target_dir), os.path.basename(f))
                zipf.write(f, arcname)    
    finally:
        if zipf:
            zipf.close()
    
    return os.path.join(target_dir, archive_name)


class Scene (object):
    
    def __init__(self):
        self.doc = c4d.documents.GetActiveDocument()
        self.export_flags = c4d.SAVEPROJECT_ASSETS
        self.export_flags += c4d.SAVEPROJECT_SCENEFILE
        self.export_flags += c4d.SAVEPROJECT_DIALOGSALLOWED
        self.export_flags += c4d.SAVEPROJECT_SHOWMISSINGASSETDIALOG
        self.export_flags += c4d.SAVEPROJECT_PROGRESSALLOWED
        self.export_flags += c4d.SAVEPROJECT_DONTTOUCHDOCUMENT
        if c4d.GetC4DVersion() >= 17000:
            self.export_flags += c4d.SAVEPROJECT_WITHCACHES
        if c4d.GetC4DVersion() >= 24000:
            self.export_flags += c4d.SAVEPROJECT_ASSETLINKS_COPY_FILEASSETS
            self.export_flags += c4d.SAVEPROJECT_ASSETLINKS_COPY_NODEASSETS

    @property
    def project_name(self):
        if not hasattr(self, '_project_name'):
            name = self.doc.GetDocumentName()
            self._project_name = os.path.splitext(name)[0]
        return self._project_name

    def check_assets(self):
        if c4d.GetC4DVersion() < 22000:
            assets = c4d.documents.GetAllAssets(self.doc, False, '')
            return assets is not None
        else:
            assets = []
            state = c4d.documents.GetAllAssetsNew(self.doc, False, '', c4d.ASSETDATA_FLAG_NONE, assets)
            if state == c4d.GETALLASSETSRESULT_OK: 
                return assets
    
    def export_project(self, target_dir):
        export_dir = os.path.join(target_dir, self.project_name)
        c4d.documents.SaveProject(self.doc, self.export_flags, export_dir)
        return export_dir


class MultiPartForm(object):
    """Accumulate the data to be used when posting a form."""

    def __init__(self):
        self.form_fields = []
        self.files = []
        self.boundary = mimetools.choose_boundary()
        return
    
    def get_content_type(self):
        return 'multipart/form-data; boundary=%s' % self.boundary

    def add_field(self, name, value):
        """Add a simple field to the form data."""
        self.form_fields.append((name, value))
        return

    def add_file(self, fieldname, filename, fileHandle, mimetype=None):
        """Add a file to be uploaded."""
        body = fileHandle.read()
        if mimetype is None:
            mimetype = mimetypes.guess_type(filename)[0] or 'application/octet-stream'
        self.files.append((fieldname, filename, mimetype, body))
        return
    
    def __str__(self):
        """Return a string representing the form data, including attached files."""
        # Build a list of lists, each containing "lines" of the
        # request.  Each part is separated by a boundary string.
        # Once the list is built, return a string where each
        # line is separated by '\r\n'.  
        parts = []
        part_boundary = '--' + self.boundary
        
        # Add the form fields
        parts.extend(
            [ part_boundary,
              'Content-Disposition: form-data; name="%s"' % name,
              '',
              value,
            ]
            for name, value in self.form_fields
            )
        
        # Add the files to upload
        parts.extend(
            [ part_boundary,
              'Content-Disposition: file; name="%s"; filename="%s"' % \
                 (field_name, filename),
              'Content-Type: %s' % content_type,
              '',
              body,
            ]
            for field_name, filename, content_type, body in self.files
            )
        
        # Flatten the list and add closing boundary marker,
        # then return CR+LF separated data
        flattened = list(itertools.chain(*parts))
        flattened.append('--' + self.boundary + '--')
        flattened.append('')
        return '\r\n'.join(flattened)


class UserData(object):
    
    @classmethod
    def get_data_filepath(cls):
        return os.path.join(get_write_dir(), 'data.dat')
    
    @classmethod
    def store(cls, username, password):
        filepath = cls.get_data_filepath()
        b64u = base64.b64encode(username)
        b64p = base64.b64encode(password)
        data = "{0}:{1}".format(b64u, b64p)
        with open(filepath, 'w+') as f:
            f.write(data)

    @classmethod            
    def retrive(cls):
        filepath = cls.get_data_filepath()
        
        if not os.path.exists(filepath):
            return '', ''
        
        with open(filepath, 'r') as f:
            data = f.read()
        b64u, b64p = data.split(':')
        username = base64.b64decode(b64u)
        password = base64.b64decode(b64p)
        return username, password
    
    @classmethod
    def delete_file(cls):
        filepath = cls.get_data_filepath()
        os.remove(filepath)


class RenderBuzzUploader(object):
    
    def __init__(self, scene, username, password):
        self.scene = scene
        self.url = "http://renderbuzz.com/api/upload"
        self.username = username
        self.password = password
    
    def archive_scene(self):
        export_path = create_tmp_dir()
        archive_dir = self.scene.export_project(export_path)
        files = get_files(archive_dir)
        archive_name = self.scene.project_name
        self.archive_path = zip_files(files, archive_dir, archive_name)
        self.archive_name = os.path.basename(self.archive_path)
    
    def create_http_form(self):
        form = MultiPartForm()
        form.add_field('username', self.username)
        form.add_field('password', self.password)
        with open(self.archive_path, 'rb') as f:
            form.add_file('file', self.archive_name, fileHandle=f)
        return form
    
    def upload_archive(self):
        form = self.create_http_form()
        body = str(form)
        
        request = urllib2.Request(self.url)
        request.add_header('Content-type', form.get_content_type())
        request.add_header('Content-length', len(body))
        request.add_data(body)
        
        try:
            msg = urllib2.urlopen(request).read()
        except urllib2.HTTPError as e:
            msg = e.read()
            raise Exception(msg)
    

class RenderBuzzDialog(gui.GeDialog):
    
    CST_LOGO = 11101
    LBL_USERNAME = 10000
    LBL_PASSWORD = 10001
    LBL_EMPTY = 10002
    LBL_STATUS = 10003
    TXT_USERNAME = 10010
    TXT_PASSWORD = 10011
    CHB_REMEMBERME = 10020
    BTN_UPLOAD = 10042
        
    def get_logo(self):
        resources_dir = get_resources_dir()
        logo_filepath = os.path.join(resources_dir, "logo.png")
        bmp = bitmaps.BaseBitmap()
        bmp.InitWith(logo_filepath)
        return bmp
        
    
    def CreateLayout(self):
        self.SetTitle('RenderBuzz Uploader')
        
        custom = self.AddCustomGui(self.CST_LOGO, c4d.CUSTOMGUI_BITMAPBUTTON,
                    "", c4d.BFH_LEFT, minw=260, minh=50)
        custom.SetImage(self.get_logo(), copybmp=True)
        
        self.AddStaticText(self.LBL_USERNAME, c4d.BFH_LEFT, name='Email address:')
        self.AddEditText(self.TXT_USERNAME, c4d.BFH_SCALEFIT)
        
        self.AddStaticText(self.LBL_PASSWORD, c4d.BFH_LEFT, name='Password:')
        self.AddEditText(self.TXT_PASSWORD, c4d.BFH_SCALEFIT,
            editflags=c4d.EDITTEXT_PASSWORD)
            
        self.AddCheckbox(self.CHB_REMEMBERME, c4d.BFH_LEFT, initw=200, inith=20, name="Remember me")
       
        self.AddStaticText(self.LBL_EMPTY, c4d.BFH_LEFT, name='', inith=40) 
        self.AddButton(self.BTN_UPLOAD, c4d.BFH_SCALEFIT, name="Upload",
                       inith=30)
        self.AddStaticText(self.LBL_STATUS, c4d.BFH_SCALEFIT, name="Welcome to RenderBuzz Uploader")
        
        self.restore_user_data()   
        
        return True
    
    def log_mssage(self, msg):
        self.SetString(self.LBL_STATUS, msg)
        c4d.DrawViews(c4d.DA_STATICBREAK)

    def upload(self):
        self.Enable(self.BTN_UPLOAD, False)
        self.log_mssage("Collecting scene data")
        scene = Scene()
        
        if not scene.check_assets():
            gui.MessageDialog("ERROR: Can't get all assets from document.")
            return False
        
        uploader = RenderBuzzUploader(scene, self.username, self.password)
        try:
            self.log_mssage("Archiving scene")
            uploader.archive_scene()
            self.log_mssage("Uploading scene")
            uploader.upload_archive()
        except Exception as e:
            msg = "ERROR: {0}".format(e)
            gui.MessageDialog(msg)
            self.log_mssage("Upload aborted")
            return False
        else:
            self.log_mssage("Log in to RenderBuzz and start rendering")
        
        self.Enable(self.BTN_UPLOAD, True)
        return True
    
    def validate_frame_rate(self):
        doc = c4d.documents.GetActiveDocument()
        render_fps = doc.GetActiveRenderData()[c4d.RDATA_FRAMERATE]
        project_fps = doc[c4d.DOCUMENT_FPS]
        return render_fps == project_fps
    
    def display_frame_rate_warning(self):
        return c4d.gui.QuestionDialog('Project frame rate and output frame rate have different values. It may cause the difference between specified frame range and actual frames being rendered. \r\nDo you want to continue with these settings?')

    def validate_user_data(self):
        self.username = self.GetString(self.TXT_USERNAME)
        self.password = self.GetString(self.TXT_PASSWORD)
        
        if not (self.username and self.password):
            gui.MessageDialog("ERROR: Email and password fields can't be empty.")
            return False
        return True
    
    def save_user_data(self):
        UserData.store(self.username, self.password)
        
    def remove_user_data(self):
        self.SetString(self.TXT_USERNAME, '')
        self.SetString(self.TXT_PASSWORD, '', flags=c4d.EDITTEXT_PASSWORD)
        self.SetBool(self.CHB_REMEMBERME, False)
        UserData.delete_file()
        
    def restore_user_data(self):
        username, password = UserData.retrive()
        if username and password:
            self.SetString(self.TXT_USERNAME, username)
            self.SetString(self.TXT_PASSWORD, password, flags=c4d.EDITTEXT_PASSWORD)
            self.SetBool(self.CHB_REMEMBERME, True)
    
    def Command(self, id, msg):
        gui.SetMousePointer(c4d.MOUSE_BUSY)
        if id not in (self.BTN_UPLOAD, self.CHB_REMEMBERME):
            return False
        
        if not self.validate_user_data():
            self.SetBool(self.CHB_REMEMBERME, False)
            return False
                
        if id == self.BTN_UPLOAD:
            if not self.validate_frame_rate() and not self.display_frame_rate_warning():
                return False
            return self.upload()
        elif id == self.CHB_REMEMBERME:
            checked = self.GetBool(self.CHB_REMEMBERME)
            if checked:
                self.save_user_data()
            else:
                self.remove_user_data()
        gui.SetMousePointer(c4d.MOUSE_NORMAL)
        return True

class RenderBuzzPlugin(plugins.CommandData):
    
    def __init__(self):
        self.data = {}
        
    def Execute(self, doc):
        dlg = RenderBuzzDialog()
        dlg.Open(c4d.DLG_TYPE_MODAL, defaultw=268, defaulth=310)
        return True
        

def main():

    production = True # set to False to test plugin in c4d without installing
    
    if production:    
        resource_dir = get_resources_dir()
        icon_filepath = os.path.join(resource_dir, "icon.tif")
        bmp = bitmaps.BaseBitmap()
        bmp.InitWith(icon_filepath)
        plugins.RegisterCommandPlugin(PLUGIN_ID, "RenderBuzz Uploader", 0,
                                      bmp, "", RenderBuzzPlugin())
    else:
        dlg = RenderBuzzDialog()
        dlg.Open(c4d.DLG_TYPE_MODAL, defaultw=268, defaulth=310)

if __name__== '__main__':
    main()
